#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"
#include "lis_solver.h"

#define LIS_SOLVERS_LEN			12
#define LIS_PRECON_TYPE_LEN		9
#define NWORK_CG				8
#define NWORK_BICG				8
#define NWORK_CGS				8
#define NWORK_BICGSTAB			8
#define NWORK_BICGSTABL			4
#define NWORK_GPBICG			15
#define NWORK_ORTHOMIN			3
#define NWORK_GMRES				4
#define NWORK_QMR				11
#define NWORK_JACOBI			3
#define NWORK_GS				3
#define NWORK_SOR				3
int lis_nwork[] ={	0, NWORK_CG     ,NWORK_BICG     ,NWORK_CGS   ,NWORK_BICGSTAB ,NWORK_BICGSTABL,
				NWORK_GPBICG ,NWORK_ORTHOMIN ,NWORK_GMRES,NWORK_QMR       ,NWORK_JACOBI,
				NWORK_GS     ,NWORK_SOR
			};
char *lis_solvername[] = {	"", "CG"          ,"BiCG"     ,"CGS"   ,"BiCGSTAB" ,"BiCGSTAB(l)",
						"GPBICG"    ,"TFQMR"  ,"orthomin" ,"gmres"       ,"Jacobi",
						"Gauss-Seidel","SOR"
			};
char *lis_preconname[] = {	"None", "Jacobi", "ILU", "SSOR", "HYBRID", "I+S", "SAINV", "SAAMG", "RIC2S"
			};

char *lis_returncode[] = {"LIS_SUCCESS", "LIS_ILL_OPTION", "LIS_BREAKDOWN", "LIS_OUT_OF_MEMORY", "LIS_MAXITER"};

LIS_VECTOR lis_residual_history = NULL;
int lis_residual_history_count = 0;

int lis_solve(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, LIS_SCALAR params[], int options[], LIS_SCALAR status[])
{
	int solvers;
	int precon_type;
	int maxiter;
	int restart;
	int output;
	int scale;
	int err;
	int i,n,np;
	size_t worklen;
	double ptimes, itimes;
	LIS_SCALAR nrm2;
	LIS_VECTOR d;
	LIS_VECTOR *work;
	LIS_VECTOR bb;
	LIS_VECTOR xx;
	int *perm;
	LIS_PRECON precon;
	LIS_MATRIX AA;
	LIS_MATRIX At;

	status[LIS_STATUS_RET]    = LIS_ILL_OPTION;
	status[LIS_STATUS_ITER]   = 0;
	status[LIS_STATUS_RESID]  = 0;
	status[LIS_STATUS_TIMES]  = 0;
	status[LIS_STATUS_PTIMES] = 0;
	status[LIS_STATUS_ITIMES] = 0;

	/* begin parameter check */
	if( A==NULL )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	if( A->matrix_type==LIS_MATRIX_NULL )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	if( A->matrix_type<=LIS_MATRIX_ASSEMBLING )
	{
		SETERR("matrix A is assembling\n");
		return LIS_FAILS;
	}
	if( b==NULL )
	{
		SETERR("vector b is undefined\n");
		return LIS_FAILS;
	}
	if( x==NULL )
	{
		SETERR("vector x is undefined\n");
		return LIS_FAILS;
	}
	n       = A->n;
	np      = A->np;
	d       = NULL;
	At      = NULL;
	xx      = NULL;
	perm    = NULL;
	d       = NULL;
	work    = NULL;

	if( n<0 )
	{
		SETERR("Dimension n(=%d) of matrix A is less than 0\n",n);
		return LIS_FAILS;
	}
	solvers = options[LIS_OPTIONS_SOLVER];
	if( solvers < 1 || solvers > LIS_SOLVERS_LEN )
	{
		SETERR("Parameter LIS_OPTIONS_SOLVER is %d (Set between 1 to %d)\n",solvers, LIS_SOLVERS_LEN);
		return LIS_FAILS;
	}
	precon_type = options[LIS_OPTIONS_PRECON];
	if( precon_type < 0 || precon_type >= LIS_PRECON_TYPE_LEN )
	{
		SETERR("Parameter LIS_OPTIONS_PRECON is %d (Set between 0 to %d)\n",precon_type, LIS_PRECON_TYPE_LEN-1);
		return LIS_FAILS;
	}
	maxiter = options[LIS_OPTIONS_MAXITER];
	if( maxiter<0 )
	{
		SETERR("Parameter LIS_OPTIONS_MAXITER(=%d) is less than 0\n",maxiter);
		return LIS_FAILS;
	}
	restart = options[LIS_OPTIONS_RESTART];
	if( (solvers==LIS_SOLVER_GMRES || solvers==LIS_SOLVER_ORTHOMIN) && restart<0 )
	{
		SETERR("Parameter LIS_OPTIONS_RESTART(=%d) is less than 0\n",restart);
		return LIS_FAILS;
	}
	else if( solvers==LIS_SOLVER_BICGSTABL && restart<1 )
	{
		SETERR("Parameter LIS_OPTIONS_ELL(=%d) is less than 1\n",restart);
		return LIS_FAILS;
	}
	output = options[LIS_OPTIONS_OUTPUT];
	scale  = options[LIS_OPTIONS_SCALE];

	if( A->matrix_type >= LIS_MATRIX_BSR && A->matrix_type <= LIS_MATRIX_VBR)
	{
		options[LIS_OPTIONS_PADDING] = _max(A->bnr,A->bnc) + _max(A->bnr,A->bnc)%2;
		worklen = (np+_max(A->bnr,A->bnc) + _max(A->bnr,A->bnc)%2)*(lis_nwork[solvers] + 1);
	}
	else
	{
		options[LIS_OPTIONS_PADDING] = 0;
		worklen = lis_nwork[solvers];
	}
	worklen = lis_nwork[solvers];
	if( solvers==LIS_SOLVER_BICGSTABL ) worklen += 2*(restart+1);
	if( solvers==LIS_SOLVER_GMRES     ) worklen +=   (restart+2);
	if( solvers==LIS_SOLVER_ORTHOMIN  ) worklen += 3*(restart+1);
	
	status[LIS_STATUS_RET] = (LIS_SCALAR)LIS_OUT_OF_MEMORY;
	err = lis_vector_duplicate(x,&xx);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	work = (LIS_VECTOR *)lis_malloc( worklen*sizeof(LIS_VECTOR) );
	for(i=0;i<worklen;i++)
	{
		lis_vector_duplicate(x,&work[i]);		
	}
	/* end parameter check */

	if( lis_residual_history )
	{
		lis_vector_destroy(lis_residual_history);
	}
	err = lis_vector_create((maxiter+1),0,A->comm,&lis_residual_history);
	if( err )
	{
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	lis_residual_history_count = 0;

	if( A->my_rank==0 )
	{
		printf("SOLVER : %s\n", lis_solvername[solvers]);
		printf("PRECON : %s\n", lis_preconname[precon_type]);
	}

	ptimes = lis_wtime();
	if( precon_type==LIS_PRECON_TYPE_IS )
	{
		err = lis_vector_duplicate(x,&d);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		lis_matrix_scaling(A,b,d,LIS_SCALE_JACOBI);
		if( solvers >= LIS_SOLVER_JACOBI )
		{
			options[LIS_OPTIONS_ISLEVEL] = 0;
		}
	}
	else if( scale )
	{
		err = lis_vector_duplicate(x,&d);
		if( err )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		if( scale==LIS_SCALE_JACOBI && solvers==LIS_SOLVER_CG )
		{
			scale = LIS_SCALE_SYMM_DIAG;
		}
		lis_matrix_scaling(A,b,d,scale);
	}
	err = lis_precon_create(A, b, precon_type, params, options, &precon);
	if( err )
	{
		return err;
	}
	if( precon_type==LIS_PRECON_TYPE_IS )
	{
		if( solvers < LIS_SOLVER_JACOBI )
		{
			AA = A;
			bb = b;
		}
		else
		{
			AA = precon->Pmat;
			bb = precon->Pb;
		}
	}
	else
	{
		AA = A;
		bb = b;
	}

	ptimes = lis_wtime() - ptimes;
	itimes = lis_wtime();
	if( solvers==LIS_SOLVER_BICG )
	{
		lis_matrix_convert(AA,&At,LIS_MATRIX_CCS,NULL);
		precon->At = At;
	}
	lis_vector_copy(x,xx);
	switch( solvers )
	{
	case LIS_SOLVER_CG:
		err = lis_cg(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_BICG:
		err = lis_bicg(bb,xx,work,params,options,AA,At,precon,status);
		break;
	case LIS_SOLVER_CGS:
		err = lis_cgs(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_BICGSTAB:
		err = lis_bicgstab(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_BICGSTABL:
		err = lis_bicgstabl(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_GPBICG:
		err = lis_gpbicg(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_ORTHOMIN:
		err = lis_orthomin(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_GMRES:
		err = lis_gmres(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_QMR:
		err = lis_qmr(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_JACOBI:
		err = lis_jacobi(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_GS:
		err = lis_gs(bb,xx,work,params,options,AA,precon,status);
		break;
	case LIS_SOLVER_SOR:
		err = lis_sor(bb,xx,work,params,options,AA,precon,status);
		break;
	}
	status[LIS_STATUS_RET] = (LIS_SCALAR)err;

	if( scale==LIS_SCALE_SYMM_DIAG && precon_type!=LIS_PRECON_TYPE_IS)
	{
		#pragma omp parallel for
		for(i=0;i<n;i++)
		{
			x->value[i] = xx->value[i]*d->value[i];
		}
	}
	else
	{
		lis_vector_copy(xx,x);
	}
	itimes = lis_wtime() - itimes - status[LIS_STATUS_PTIMES];
	ptimes += status[LIS_STATUS_PTIMES];
	status[LIS_STATUS_TIMES]  = ptimes + itimes;
	status[LIS_STATUS_PTIMES] = ptimes;
	status[LIS_STATUS_ITIMES] = itimes;

	lis_matvec(A,xx,work[0]);
	lis_vector_xpay(b,-1,work[0]);
	if( scale==LIS_SCALE_SYMM_DIAG && precon_type!=LIS_PRECON_TYPE_IS)
	{
		#pragma omp parallel for
		for(i=0;i<n;i++)
		{
			work[0]->value[i] = work[0]->value[i]/d->value[i];
		}
	}
	lis_vector_nrm2(work[0],&nrm2);
	status[LIS_STATUS_RESID] = nrm2;
	if( A->my_rank==0 )
	{
		if( err )
		{
			printf("lis_solve is %s(code=%d)\n",lis_returncode[err],err);
		}
		else
		{
			printf("lis_solve is normal end\n");
		}
	}

	for(i=0;i<worklen;i++)
	{
		lis_vector_destroy(work[i]);
	}
	lis_free(work);
	lis_vector_destroy(d);
	lis_vector_destroy(xx);
	lis_precon_destroy(precon);
	lis_matrix_destroy(At);
	return LIS_SUCCESS;
}

int lis_get_initial_residual(LIS_MATRIX A, LIS_PRECON M, LIS_VECTOR x, LIS_VECTOR b, LIS_VECTOR r, LIS_VECTOR t, LIS_SCALAR tol, LIS_SCALAR *bnrm2, LIS_SCALAR status[])
{
	int n;
	LIS_SCALAR nrm2,dot;
	LIS_VECTOR p;

	n = A->n;
	/* ||b|| */
	lis_vector_nrm2(b,bnrm2);
	if( *bnrm2 == 0.0 )
	{
		*bnrm2 = 1.0;
	}
	else
	{
		*bnrm2 = 1.0 / *bnrm2;
	}
	/* Initial Residual */
	if( M==NULL )
	{
		p = r;
	}
	else
	{
		p = t;
	}

	lis_vector_dot(x,x,&dot);
	if( dot != 0.0 )
	{
		lis_matvec(A,x,p);      /* p = Ax    */
		lis_vector_xpay(b,-1,p);     /* p = b - p */
		lis_vector_nrm2(p,&nrm2);
		nrm2 = nrm2 * *bnrm2;
		if( nrm2 <= tol )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)1;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_FAILS;
		}
	}
	else
	{
		lis_vector_copy(b,p);
	}
	if( M!=NULL )
	{
		/* r = M^-1 * p */
		lis_psolve(M, p, r);
	}

	return LIS_SUCCESS;
}

void lis_get_solve_status(LIS_SCALAR status[], int *returnval, int *iter, LIS_SCALAR *residual, LIS_SCALAR *times)
{
	*returnval = (int)status[LIS_STATUS_RET];
	*iter      = (int)status[LIS_STATUS_ITER];
	*residual  = status[LIS_STATUS_RESID];
	*times     = status[LIS_STATUS_TIMES];
	return;
}

void lis_set_residual_history(LIS_SCALAR res)
{
	lis_residual_history->value[lis_residual_history_count++] = res;
	return;
}

int lis_get_residual_history(LIS_VECTOR v)
{
	int		i,n;

	n = _min(v->n,lis_residual_history_count);
	for(i=0;i<n;i++)
	{
		v->value[i] = lis_residual_history->value[i];
	}
	return LIS_SUCCESS;
}

int lis_output_residual_history(char *filename)
{
	int		i,my_rank,err;
	FILE	*file;

#ifdef _MPI
	my_rank = lis_residual_history->my_rank;
	err = 0;
	if( my_rank==0 )
	{
		file = fopen(filename, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", filename);
			err = 1;
		}
		else
		{
			for(i=0;i<lis_residual_history_count;i++)
			{
				fprintf(file, "%e\n", lis_residual_history->value[i]);
			}
			fclose(file);
		}
	}
	MPI_Barrier(lis_residual_history->comm);
	return LIS_SUCCESS;
#else
	file = fopen(filename, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", filename);
		return LIS_FAILS;
	}
	for(i=0;i<lis_residual_history_count;i++)
	{
		fprintf(file, "%e\n", lis_residual_history->value[i]);
	}
	fclose(file);
	return LIS_SUCCESS;
#endif
}

int lis_get_solvername(int sol, char *name)
{
	if( sol < 1 || sol > LIS_SOLVERS_LEN )
	{
		name = NULL;
		return LIS_FAILS;
	}
	strcpy(name,lis_solvername[sol]);

	return LIS_SUCCESS;
}

